from snitch.parsers.openapi_parser import OpenApiParser
from snitch.parsers.request_model import Request
import unittest
from unittest.mock import patch, mock_open
from yaml.parser import ParserError


class TestOpenApiParser(unittest.TestCase):
    def setUp(self):
        self.data = '''openapi: 3.0.0
info:
  title: FarmWeather
  version: 1.0.0
servers:
  - url: '{{apiDomain}}'
paths:
  /users:
    get:
      tags:
        - default
      summary: Test API
      parameters:
        - name: Authorization
          in: header
          schema:
            type: string
          example: '{{accessToken}}'
        - name: x-api-key
          in: header
          schema:
            type: string
          example: '{{apiKey}}'
        - name: uid
          in: query
          schema:
            type: integer
          example: '123'
      responses:
        '200':
          description: Successful response
          content:
            application/json: {}'''

    def test_OpenApiParser(self):
        with patch("builtins.open", mock_open(read_data=self.data)) as mock_file:
            reqs = OpenApiParser('')

        self.assertEqual(len(reqs.requests), 1)
        self.assertEqual(type(reqs.requests[0]), Request)
        self.assertEqual(reqs.requests[0].url, '{{apiDomain}}/users?uid=123')

    def test_OpenApiParser_error(self):
        with patch("builtins.open", mock_open(read_data=self.data+',')) as mock_file:
            with self.assertRaises(ParserError):
                reqs = OpenApiParser('')



from __future__ import annotations

from dataclasses import dataclass
from typing import Union, Dict


@dataclass
class Request:
    method: str
    url: str
    headers: Dict
    body: Dict = None
    name: str = ''


@dataclass
class RequestHeader:
    key: str
    value: Union[str, int, float]



import asyncclick as click
from datetime import datetime
from typing import Dict


class ReportBuilder:
    '''
    ReportBuilder builds formatted report strings
    '''

    def __init__(self, data: Dict):
        self.__data = data

    @property
    def header(self) -> str:
        return f"Datetime: {datetime.now()}{self.newline}Total APIs checked: {self.__data['total']}{self.newline}" + \
            f"Success: {self.__data['success']}{self.newline}Errors: {self.__data['errors']}{self.newline}" + \
            f"Success rate: {self.__data['success'] / self.__data['total'] * 100:.2f}%"

    @property
    def content(self) -> str:
        s = ''
        if 'responses' in self.__data:
            for r in self.__data['responses']:
                s += f"{r['message']}{self.newline * 2}"
            return s
        else:
            return ''

    @property
    def newline(self) -> str:
        return '\n'

    @property
    def divider(self) -> str:
        return f"{'-' * 50}"

    def print_newline(self) -> None:
        click.echo(self.newline)

    def print_divider(self) -> None:
        click.echo(self.divider)

    def print_content(self) -> None:
        for r in self.__data['responses']:
            if not r['error']:
                click.echo(r['message'])
            else:
                click.echo(click.style(
                    r['message'], fg='red'))
            self.print_newline()

    def print_summary(self) -> None:
        click.echo(click.style(
            f"Datetime: {datetime.now()}", fg='green'))
        click.echo(click.style(
            f"Total APIs checked: {self.__data['total']}", fg='green'))
        click.echo(click.style(
            f"Success: {self.__data['success']}", fg='green'))
        click.echo(click.style(f"Errors: {self.__data['errors']}", fg='red'))
        click.echo(click.style(
            f"Success rate: {self.__data['success'] / self.__data['total'] * 100:.2f}%", fg='green'))



from snitch.parsers.config_parser import ConfigParser
from snitch.parsers.parser_exceptions import InvalidPostmanCollectionVersion, InvalidOpenApiVersion
import unittest
import json


class TestConfigParser(unittest.TestCase):
    def setUp(self):
        self.__data = '''{
            "postmanCollection": {
                "version": "2.1",
                "filePath": "absolute_path_to_the_postman_collection_json_file",
                "metadata": {
                    "{{restUrl}}": "https://your_api_domain",
                    "{{accessToken}}": "the access token string",
                    "{{apiKey}}": "api key string"
                },
                "header": {
                    "Authorization": "{{accessToken}}",
                    "x-api-key": "{{apiKey}}"
                }
            },
            "openApi": {
                "version": "3.0.0",
                "filePath": "absolute_path_to_the_open_api_yaml_file",
                "metadata": {
                    "{{restUrl}}": "https://your_api_domain",
                    "{{accessToken}}": "the access token string",
                    "{{apiKey}}": "api key string"
                },
                "header": {
                    "Authorization": "{{accessToken}}",
                    "x-api-key": "{{apiKey}}"
                }
            }
        }'''

    def test_ConfigParser(self):
        config = ConfigParser(self.__data)

        self.assertTrue(config.has_postman_collection)
        self.assertTrue(config.has_openapi)
        self.assertEqual(config.collection_metadata, {
            "{{restUrl}}": "https://your_api_domain",
            "{{accessToken}}": "the access token string",
            "{{apiKey}}": "api key string"
        })
        self.assertEqual(config.openapi_metadata, {
            "{{restUrl}}": "https://your_api_domain",
            "{{accessToken}}": "the access token string",
            "{{apiKey}}": "api key string"
        })
        self.assertEqual(config.collection_file_path,
                         "absolute_path_to_the_postman_collection_json_file")
        self.assertEqual(config.openapi_file_path,
                         "absolute_path_to_the_open_api_yaml_file")

    def test_ConfigParser_InvalidPostmanCollectionVersion(self):
        tmp = json.loads(self.__data)
        tmp['postmanCollection']['version'] = '1.0'
        tmp['openApi']['version'] = '3.0.0'
        self.__data = json.dumps(tmp)

        with self.assertRaises(InvalidPostmanCollectionVersion):
            ConfigParser(self.__data)

    def test_ConfigParser_InvalidOpenApiVersion(self):
        tmp = json.loads(self.__data)
        tmp['postmanCollection']['version'] = '2.1'
        tmp['openApi']['version'] = '2.0.0'
        self.__data = json.dumps(tmp)

        with self.assertRaises(InvalidOpenApiVersion):
            ConfigParser(self.__data)



from snitch.task_runners.idempotency_check import run_idempotency_check
from snitch.parsers.request_model import Request
from snitch.logger import LogItem
import unittest
from unittest.mock import patch, Mock, MagicMock


class TestIdempotencyCheck(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.data = [Request('GET', 'https://example.com',
                             {'content-type', 'application/json'}, None, 'Test API')]
        self.request = Mock()
        self.request.url = 'https://example.com'
        self.request.headers = {'content-type': 'application/json'}
        self.request.body = None
        self.request.method = 'GET'
        self.request.name = 'Test API'

        self.response = Mock()
        self.response.status_code = 201
        self.response.text = 'content'

    async def test_run_idempotency_check_get(self):
        with patch("requests.get") as request_mock:
            request_mock.return_value = self.response
            res = await run_idempotency_check(self.request)

        self.assertEqual(res['error'], False)

    async def test_run_idempotency_check_post(self):
        with patch("requests.post") as async_mock:
            self.request.method = 'POST'
            async_mock.return_value = self.response

            res = await run_idempotency_check(self.request)

        self.assertEqual(res['error'], False)

    async def test_run_idempotency_check_put(self):
        with patch("requests.put") as async_mock:
            self.request.body = {}
            self.request.method = 'PUT'
            async_mock.return_value = self.response

            res = await run_idempotency_check(self.request)

        self.assertEqual(res['error'], False)

    async def test_run_idempotency_check_delete_error(self):
        with patch("requests.delete") as async_mock:
            self.request.body = None
            self.request.method = 'DELETE'
            self.response.status_code = 400
            async_mock.return_value = self.response
            res = await run_idempotency_check(self.request)

        self.assertEqual(res['error'], True)



from snitch.task_runners.health_check import run_health_check, request
from snitch.parsers.request_model import Request
from snitch.logger import LogItem
import unittest
import aiohttp
from unittest.mock import patch, AsyncMock, Mock, MagicMock


class TestHealthCheck(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.data = [Request('GET', 'https://example.com',
                             {'content-type', 'application/json'}, None, 'Test API')]
        self.request = Mock()
        self.request.url = 'https://example.com'
        self.request.headers = {'content-type': 'application/json'}
        self.request.body = None
        self.request.method = 'GET'
        self.request.name = 'Test API'

        self.response = Mock()
        self.response.status = 201
        self.response.content = 'content'
        self.response.text = lambda: 'text'

    async def test_run_health_check_get(self):
        with patch("snitch.task_runners.health_check.get_all_requests", new_callable=AsyncMock) as async_mock:
            async_mock.return_value = [
                LogItem(False, 1, '', 'Test API')]
            res = await run_health_check(self.data)

            self.assertEqual(res['total'], 1)
            self.assertEqual(res['success'], 1)

    async def test_request_get(self):
        with patch("aiohttp.ClientSession.get", new_callable=AsyncMock) as async_mock:
            async_mock.return_value = self.response

            session = aiohttp.ClientSession
            session.get = MagicMock()
            session.get.return_value.__aenter__.return_value.status = 200
            session.get.return_value.__aenter__.return_value.text.return_value = 'test content'
            res = await request(session, self.request)

        self.assertEqual(res.has_err, False)
        self.assertEqual(res.name, 'Test API')

    async def test_request_post(self):
        with patch("aiohttp.ClientSession.post", new_callable=AsyncMock) as async_mock:
            self.request.method = 'POST'
            async_mock.return_value = self.response

            session = aiohttp.ClientSession
            session.post = MagicMock()
            session.post.return_value.__aenter__.return_value.status = 200
            session.post.return_value.__aenter__.return_value.text.return_value = 'test content'
            res = await request(session, self.request)

        self.assertEqual(res.has_err, False)
        self.assertEqual(res.name, 'Test API')

    async def test_request_put(self):
        with patch("aiohttp.ClientSession.put", new_callable=AsyncMock) as async_mock:
            self.request.body = {}
            self.request.method = 'PUT'

            async_mock.return_value = self.response

            session = aiohttp.ClientSession
            session.put = MagicMock()
            session.put.return_value.__aenter__.return_value.status = 200
            session.put.return_value.__aenter__.return_value.text.return_value = 'test content'
            res = await request(session, self.request)

        self.assertEqual(res.has_err, False)
        self.assertEqual(res.name, 'Test API')

    async def test_request_delete_error(self):
        with patch("aiohttp.ClientSession.delete", new_callable=AsyncMock) as async_mock:
            self.request.body = None
            self.request.method = 'DELETE'
            async_mock.return_value = self.response

            session = aiohttp.ClientSession
            session.delete = MagicMock()
            session.delete.return_value.__aenter__.return_value.status = 400
            session.delete.return_value.__aenter__.return_value.text.return_value = 'test content'
            res = await request(session, self.request)

        self.assertEqual(res.has_err, True)
        self.assertEqual(res.name, 'Test API')



class InvalidPostmanCollectionVersion(Exception):
    def __init__(self, message) -> None:
        super().__init__(message)
        self.message = message


class InvalidOpenApiVersion(Exception):
    def __init__(self, message) -> None:
        super().__init__(message)
        self.message = message



from snitch.parsers.postman_parser import PostmanCollectionParser
from snitch.parsers.request_model import Request
import unittest
from unittest.mock import patch, mock_open
from json.decoder import JSONDecodeError


class TestPostmanCollectionParser(unittest.TestCase):
    def setUp(self):
        self.data = '''{"info": {"_postman_id": "8d7c5957-b22d-40f1-a155-9116a1a92442","name": "test",
                            "schema": "https://schema.getpostman.com/json/collection/v2.1.0/collection.json",
                            "_exporter_id": "6670289"}, "item": [{"name": "Test endpoint","request": {"method": "GET","header":
                            [{"key": "Authorization","type": "text","value": "{{accessToken}}"},{
                                "key": "x-api-key","type": "text","value": "{{apiKey}}"}],
                                    "url": {"raw": "{{apiDomain}}/users?uid=123","host": ["{{apiDomain}}"],"path": ["users"],"query":
                                    [{"key": "uid","value": "123"}]}},"response": []}]}'''

    def test_PostmanCollectionParser(self):
        with patch("builtins.open", mock_open(read_data=self.data)) as mock_file:
            reqs = PostmanCollectionParser('')

        self.assertEqual(len(reqs.requests), 1)
        self.assertEqual(type(reqs.requests[0]), Request)
        self.assertEqual(reqs.requests[0].url, '{{apiDomain}}/users?uid=123')

    def test_PostmanCollectionParser_error(self):
        with patch("builtins.open", mock_open(read_data=self.data+',')) as mock_file:
            with self.assertRaises(JSONDecodeError):
                reqs = PostmanCollectionParser('')



from snitch.task_runners.report_builder import ReportBuilder
import unittest
from unittest.mock import patch


class TestReportBuilder(unittest.TestCase):
    def setUp(self):
        self.data = {
            'total': 100,
            'success': 50,
            'errors': 50,
            'responses': [{'message': 'test message'}]
        }
        self.report_builder = ReportBuilder(self.data)

    def test_header(self):
        with patch("snitch.task_runners.report_builder.datetime") as mock:
            mock.now.return_value = 123
            self.assertEqual(self.report_builder.header,
                             f"Datetime: 123{self.report_builder.newline}Total APIs checked: 100{self.report_builder.newline}Success: 50{self.report_builder.newline}Errors: 50{self.report_builder.newline}Success rate: 50.00%")

    def test_conetnt(self):
        self.assertEqual(self.report_builder.content,
                         f"test message{self.report_builder.newline*2}")

    def test_other(self):
        self.assertEqual(self.report_builder.newline, '\n')
        self.assertEqual(self.report_builder.divider, f"{'-' * 50}")



from time import time
import requests
from typing import Dict
from ..parsers.request_model import Request
from ..logger import LogItem


async def run_idempotency_check(req: Request) -> Dict[str, str]:
    try:
        start = time()
        if req.method == 'GET':
            res1 = requests.get(req.url, headers=req.headers)
            if res1.status_code >= 400:
                raise Exception()
            res2 = requests.get(req.url, headers=req.headers)
            if res2.status_code >= 400:
                raise Exception()
            is_idempotent = res1.text == res2.text
        elif req.method == 'POST':
            res1 = requests.post(
                req.url, headers=req.headers, json=req.body)
            if res1.status_code >= 400:
                raise Exception()
            res2 = requests.post(
                req.url, headers=req.headers, json=req.body)
            if res2.status_code >= 400:
                raise Exception()
            is_idempotent = res1.text == res2.text
        elif req.method == 'PUT':
            res1 = requests.put(
                req.url, headers=req.headers, json=req.body)
            if res1.status_code >= 400:
                raise Exception()
            res2 = requests.put(
                req.url, headers=req.headers, json=req.body)
            if res2.status_code >= 400:
                raise Exception()
            is_idempotent = res1.text == res2.text
        elif req.method == 'DELETE':
            res1 = requests.delete(req.url, headers=req.headers)
            if res1.status_code >= 400:
                raise Exception()
            res2 = requests.delete(req.url, headers=req.headers)
            if res2.status_code >= 400:
                raise Exception()
            is_idempotent = res1.text == res2.text
        end = time()

        log = LogItem(False, end - start,
                      f'{is_idempotent}', req.name)

        return {'error': log.has_err,
                'message':
                f'Name: {log.name}\nError: {log.has_err}\nLatency: {log.run_time}s\nIdempotent: {log.message}'}

    except Exception as e:
        end = time()
        log = LogItem(True, end - start,
                      f'Untestable', req.name)

        return {'error': log.has_err,
                'message':
                f'Name: {log.name}\nError: {log.has_err}\nLatency: {log.run_time}s\nIdempotent: {log.message}'}



from snitch.parsers.request_model import Request, RequestHeader
import unittest


class TestRequestModel(unittest.TestCase):
    def test_Request(self):
        r = Request('POST', 'https://example.com',
                    {'authorization': 'token', 'x-api-key': 'key'}, None, 'test api')

        self.assertEqual(r.name, 'test api')
        self.assertEqual(r.url, 'https://example.com')
        self.assertEqual(r.method, 'POST')
        self.assertEqual(r.body, None)
        self.assertEqual(
            r.headers, {'authorization': 'token', 'x-api-key': 'key'})

    def test_RequestHeader(self):
        r = RequestHeader('content-type', 'application/json')

        self.assertEqual(r.key, 'content-type')
        self.assertEqual(r.value, 'application/json')



from snitch.task_runners.task import TaskType, Task, TaskQueue
import unittest


class TestTaskType(unittest.IsolatedAsyncioTestCase):
    def test_TaskType(self):
        self.assertEqual(TaskType.HEALTH_CHECK.value, 'hc')
        self.assertEqual(TaskType.IDEMPOTENCY_CHECK.value, 'id')

    async def test_TaskQueue(self):
        async def dummy_func(a):
            return a

        task_queue = TaskQueue()
        task_queue.add_task(Task(dummy_func, 1))
        self.assertEqual(len(task_queue), 1)

        task = task_queue.pop_task()

        self.assertEqual(await task.task(task.args), 1)



import asyncclick as click
import asyncio
import os
import uuid
from datetime import date
from typing import List, Optional
from .parsers.postman_parser import PostmanCollectionParser
from .parsers.openapi_parser import OpenApiParser
from .parsers.config_parser import ConfigParser
from .task_runners.health_check import run_health_check
from .task_runners.idempotency_check import run_idempotency_check
from .task_runners.report_builder import ReportBuilder
from .scripts.generate_config_json_template import generate_config_json_template
from .task_runners.task import TaskType, TaskQueue, Task
from .parsers.request_model import Request


@click.command()
@click.option('-i', '--init', help='To create a new config.json template file, you need to provide a directory to store your config file', default='')
@click.option('-o', '--output', help='To specify an existing directory that stores task results on your device', default='')
@click.option('-p', '--path', help='To specify a path where stores the config JSON file')
@click.option('-t', '--task', help='To run a specific task, snitch will run all tasks if this flag is not provided', default='')
@click.option('-v', '--verbose', is_flag=True, help='To print API responses if -v or --verbose flag is provided')
async def run(path, output: Optional[str], init: Optional[str], task: Optional[str], verbose: Optional[bool]) -> None:
    try:
        if init:
            generate_config_json_template(init)
        else:
            if not path:
                raise Exception('Error. No -p or --path flag specified.')
            with open(path, 'r') as f:
                click.echo(click.style(
                    'Analyzing the configuration...\n', fg='yellow'))
                config = ConfigParser(f.read())
                reqs = []

                if config.has_postman_collection:
                    pp = PostmanCollectionParser(
                        config.collection_file_path, config.collection_metadata)

                    reqs.extend(pp.requests)
                    # run health check and API validation, and handle the result

                if config.has_openapi:
                    op = OpenApiParser(
                        config.openapi_file_path, config.openapi_metadata)
                    reqs.extend(op.requests)

                if reqs:
                    task_queue = TaskQueue()
                    if task:
                        match task:
                            case TaskType.HEALTH_CHECK.value:
                                task_queue.add_task(
                                    Task(run_health_task, (output, reqs, verbose)))
                            case TaskType.IDEMPOTENCY_CHECK.value:
                                task_queue.add_task(
                                    Task(run_idempotency_task, (output, reqs)))
                            case _: raise Exception(f'Error. \'{task}\' is an invalid task flag.')
                    else:
                        task_queue.add_task(
                            Task(run_health_task, (output, reqs, verbose)))
                        task_queue.add_task(
                            Task(run_idempotency_task, (output, reqs)))

                    while task_queue:
                        t = task_queue.pop_task()
                        await t.task(*t.args)
                else:
                    raise Exception('Error. No APIs available to check.')

    except Exception as e:
        click.echo(click.style(e, fg='red'))


async def run_health_task(output: str, requests: List[Request], verbose: bool) -> None:
    click.echo(click.style(
        'Running Health check ...', fg='yellow'))
    res = await run_health_check(requests, verbose)
    report_builder = ReportBuilder(res)
    report_builder.print_divider()
    report_builder.print_newline()
    report_builder.print_content()
    report_builder.print_divider()
    report_builder.print_summary()

    if output:
        if os.path.exists(output):
            f = open(os.path.join(
                output, generate_file_name('health_check')), 'w')

            f.write(report_builder.header)
            f.write(report_builder.newline)
            f.write(report_builder.divider)
            f.write(report_builder.newline * 2)
            f.write(report_builder.content)
            f.close()
        else:
            click.echo(
                click.style(
                    'Error: The output destination directory doesn\'t exist.',
                    fg='red'))


async def run_idempotency_task(output: str, requests: List[Request]) -> None:
    click.echo(click.style(
        '\nRunning Idempotency check, it will take longer...', fg='yellow'))
    idem_responses = []
    for re in requests:
        res = await run_idempotency_check(re)
        idem_responses.append(res)

    res = {}

    total = len(idem_responses)

    success = len(
        [i for i in idem_responses if not i['error']])
    res['total'] = total
    res['success'] = success
    res['errors'] = total - success
    res['responses'] = idem_responses

    report_builder = ReportBuilder(res)

    report_builder.print_divider()
    report_builder.print_newline()
    report_builder.print_content()
    report_builder.print_divider()
    report_builder.print_summary()

    if output:
        if os.path.exists(output):
            f = open(os.path.join(
                output, generate_file_name('idempotency')), 'w')

            f.write(report_builder.header)
            f.write(report_builder.newline)
            f.write(report_builder.divider)
            f.write(report_builder.newline * 2)
            f.write(report_builder.content)
            f.close()
        else:
            click.echo(
                click.style(
                    'Error: The output destination directory doesn\'t exist.',
                    fg='red'))


def generate_file_name(type: str) -> str:
    return f'{type}_{date.today().strftime("%Y_%m_%d")}_{uuid.uuid4()}.log'


if __name__ == '__main__':
    asyncio.run(run())



import json
from typing import Union, Dict
from .parser_exceptions import InvalidOpenApiVersion, InvalidPostmanCollectionVersion
from ..typed import Number


class ConfigParser:
    # It takes a json string as input during object instantiation
    def __init__(self, data: str) -> None:
        self.__postman_collection = {}
        self.__openapi = {}

        try:
            config = json.loads(data)
            if 'postmanCollection' in config:
                if int(config['postmanCollection']['version'].split('.')[0]) < 2:
                    raise InvalidPostmanCollectionVersion(
                        'Error. snitch requires Postman collection version >= 2.0')
                self.__postman_collection['file_path'] = config['postmanCollection']['filePath']
                self.__postman_collection['metadata'] = config['postmanCollection']['metadata']

            if 'openApi' in config:
                if int(config['openApi']['version'].split('.')[0]) < 3:
                    raise InvalidOpenApiVersion(
                        'Error. snitch requires Open API version >= 3.0.0')
                self.__openapi['file_path'] = config['openApi']['filePath']
                self.__openapi['metadata'] = config['openApi']['metadata']
        except Exception as e:
            raise e

    @property
    def has_postman_collection(self) -> bool:
        return True if self.__postman_collection else False

    @property
    def collection_metadata(self) -> Union[Dict[str, Union[str, Number]], None]:
        return self.__postman_collection.get('metadata', None)

    @property
    def collection_file_path(self) -> Union[str, None]:
        return self.__postman_collection.get('file_path', None)

    @property
    def has_openapi(self) -> bool:
        return True if self.__openapi else False

    @property
    def openapi_metadata(self) -> Union[Dict[str, Union[str, Number]], None]:
        return self.__openapi.get('metadata', None)

    @property
    def openapi_file_path(self) -> Union[str, None]:
        return self.__openapi.get('file_path', None)



from snitch.parsers.parser_exceptions import InvalidPostmanCollectionVersion, InvalidOpenApiVersion
import unittest


class TestExceptions(unittest.TestCase):
    def test_InvalidPostmanCollectionVersion(self):
        err_msg = 'Error. snitch requires Postman collection version >= 2.0'
        e = InvalidPostmanCollectionVersion(
            err_msg)

        self.assertEqual(e.message, err_msg)

        try:
            raise e
        except Exception as err:
            self.assertEqual(e.message, err_msg)

        err_msg = 'Error. snitch requires Open API version >= 3.0.0'
        e = InvalidOpenApiVersion(
            err_msg)

        self.assertEqual(e.message, err_msg)

        try:
            raise e
        except Exception as err:
            self.assertEqual(e.message, err_msg)



import json
from json.decoder import JSONDecodeError
from typing import List, Dict
from .request_model import Request, RequestHeader


class PostmanCollectionParser():
    def __init__(self, path, metadata={}) -> None:
        self.__requests = []
        try:
            with open(path, 'r') as f:
                data = json.loads(f.read())
                items = self.__flatten_paths(data['item'])

                for i in items:
                    # transform the header entries from list to dict
                    headers = {}

                    # edge case, item field in item
                    for h in i['request']['header']:
                        if h['value'] in metadata:
                            h['value'] = metadata[h['value']]
                        hi = RequestHeader(
                            h['key'].lower(), h['value'])
                        headers[hi.key] = hi.value

                    # edge case, if method is POST but content-type is missing
                    if i['request']['method'] in set(['POST', 'PUT']):
                        headers['content-type'] = 'application/json'

                    # replacing the placeholders in url
                    if type(i['request']['url']) == str:
                        url_str = i['request']['url']
                        for h in metadata:
                            url_str = url_str.replace(h, metadata[h])
                    else:
                        url_str = i['request']['url']['raw']

                        for h in i['request']['url']['host']:
                            if h in metadata:
                                url_str = url_str.replace(h, metadata[h])

                    # body payload could be malformatted
                    # TODO: support content-type other than application/json
                    body = None
                    try:
                        if 'body' in i['request'] and i['request']['body']['mode'] == 'raw':
                            body = json.loads(
                                i['request']['body']['raw'].strip())

                            while type(body) == str:
                                body = json.loads(body)
                    except Exception as e:
                        body = None

                    req = Request(
                        i['request']['method'], url_str, headers, body, i['name'])

                    self.__requests.append(req)
        except JSONDecodeError as e:
            raise e  # TODO: handle this differently
        except FileNotFoundError as e:
            raise e

    def __flatten_paths(self, paths: List[Dict]) -> List[Dict]:
        items = []

        def dfs(item) -> None:
            if 'item' not in item:
                items.append(item)
                return

            for i in item['item']:
                dfs(i)

        for p in paths:
            dfs(p)

        return items

    @property
    def requests(self) -> List[Request]:
        return self.__requests



import aiohttp
import asyncio
from time import time
from typing import Dict, List, Any
from ..logger import LogItem
from ..parsers.request_model import Request
from ..typed import Session


async def run_health_check(requests: List[Request], verbose: bool = False) -> Dict[str, Any]:
    async with aiohttp.ClientSession() as s:
        res = await get_all_requests(s, requests)

    status = {}
    status['total'] = len(res)
    status['errors'] = len([i for i in res if i.has_err])
    status['success'] = status['total'] - status['errors']
    status['responses'] = [
        {'error': i.has_err,
         'message':
         f'Name: {i.name}\nError: {i.has_err}\nLatency: {i.run_time}s' +
            (f'\nResponse: {i.message}' if verbose else '')}
        for i in res]

    return status


async def request(session: Session, request: Request) -> LogItem:
    try:
        start = time()
        if request.method == 'GET':
            async with session.get(request.url, headers=request.headers) as response:
                if response.status >= 400:
                    raise Exception(f'{response.status} - {response.content}')
                msg = await response.text()
        elif request.method == 'POST':
            async with session.post(request.url, headers=request.headers, json=request.body) as response:
                if response.status >= 400:
                    raise Exception(f'{response.status} - {response.content}')
                msg = await response.text()
        elif request.method == 'PUT':
            async with session.put(request.url, headers=request.headers, json=request.body) as response:
                if response.status >= 400:
                    raise Exception(f'{response.status} - {response.content}')
                msg = await response.text()
        elif request.method == 'DELETE':
            async with session.delete(request.url, headers=request.headers) as response:
                if response.status >= 400:
                    raise Exception(f'{response.status} - {response.content}')
                msg = await response.text()
        end = time()

        return LogItem(False, end - start, msg, request.name)
    except Exception as e:
        end = time()
        return LogItem(True, end - start, e, request.name)


async def get_all_requests(session: Session, requests: List[Request]) -> List[LogItem]:
    try:
        tasks = []
        for req in requests:
            tasks.append(asyncio.create_task(request(session, req)))

        result = await asyncio.gather(*tasks)

        return result
    except Exception as e:
        raise e



import json


def generate_config_json_template(file: str) -> None:
    try:
        with open(file, 'w') as f:
            content = {"postmanCollection": {"version": "", "filePath": "", "metadata": {
            }}, "openApi": {"version": "", "filePath": "", "metadata": {}}}

            json.dump(content, f, indent=2)
    except Exception as e:
        raise e



import yaml
import json
from json.decoder import JSONDecodeError
from typing import List, Dict
from .request_model import Request, RequestHeader


class OpenApiParser:
    def __init__(self, path, metadata: Dict = {}) -> None:
        self.__requests = []
        self.__metadata = metadata
        try:
            with open(path, 'r') as f:
                data = yaml.load(f, Loader=yaml.FullLoader)
                servers = [i['url'] for i in data['servers']]

                for s in servers:
                    for k, v in data['paths'].items():
                        method = self.__extract_method(v)
                        headers = self.__extract_header_items(
                            v[method.lower()]['parameters'])
                        queries = []
                        for q in [itm for itm in v[method.lower()]['parameters'] if itm['in'] == 'query']:
                            if q['example'] in self.__metadata:
                                q['example'] = self.__metadata[q['example']]
                            queries.append(f"{q['name']}={q['example']}")

                        if method in set(['POST', 'PUT']):
                            headers['content-type'] = 'application/json'

                        body = None

                        if 'requestBody' in v[method.lower()]:
                            if 'application/json' in v[method.lower()]['requestBody']['content']:
                                body = v[method.lower(
                                )]['requestBody']['content']['application/json']['schema']['example']
                            elif '*/*' in v[method.lower()]['requestBody']['content']:
                                body = v[method.lower(
                                )]['requestBody']['content']['*/*']['schema']['example']
                        while type(body) == str:
                            body = json.loads(body)

                        for key, val in self.__metadata.items():
                            s = s.replace(key, val)

                        req = Request(method, f"{s + k}{'?' + '&'.join(queries) if queries else ''}", headers,
                                      body, v[method.lower()]['summary'])

                        self.__requests.append(req)
        except JSONDecodeError as e:
            raise e  # TODO: handle this differently
        except FileNotFoundError as e:
            raise e

    def __extract_method(self, obj) -> str:
        if 'get' in obj:
            return 'GET'
        elif 'post' in obj:
            return 'POST'
        elif 'put' in obj:
            return 'PUT'
        elif 'delete' in obj:
            return 'DELETE'

    def __extract_header_items(self, arr) -> List[Dict[str, str]]:
        headers = {}
        for p in arr:
            if p['in'] == 'header':
                if p['example'] in self.__metadata:
                    p['example'] = self.__metadata[p['example']]
                h = RequestHeader(p['name'].lower(), p['example'])
                headers[h.key] = h.value

        return headers

    @property
    def requests(self):
        return self.__requests



from __future__ import annotations
from enum import Enum
from collections import deque
from typing import Callable, Tuple, Awaitable, Any
from dataclasses import dataclass


class TaskType(Enum):
    HEALTH_CHECK = 'hc'
    IDEMPOTENCY_CHECK = 'id'


class TaskQueue:
    def __init__(self) -> None:
        self.__queue = deque([])

    def add_task(self, task: Task) -> None:
        self.__queue.append(task)

    def pop_task(self) -> Task:
        return self.__queue.popleft()

    def __len__(self) -> int:
        return len(self.__queue)


@dataclass
class Task:
    task: Callable[[Tuple], Awaitable[Any]]
    args: Tuple
